unit TermInterpreter;

interface

function TermWert(s: String; x: Double): Double;
function IsOkay: Boolean;
function GetLastErrorMsg: String;
function GetLastErrorPos: Integer;

implementation

uses math;

var LastErrorMsg : String;
    LastErrorPos : Integer;
    Okay         : Boolean;
    const_e,
    arg_x        : Double;


{ Interne Hilfsfunktionen }

function GetFirstChar(var s: String): Char;
  begin
  While (Length(s) > 0) and (s[1] in [' ']) do
    Delete(s, 1, 1);             { killt fhrende Leerzeichen        }
  If Length(s) > 0 then
    Result := s[1]               { gibt das erste Zeichen zurck.... }
  else
    Result := #0;                { ... oder bei leerem String ein #0 }
  end;

function ClosingBracket(c: Char): Char;
  begin
  Case c of
    '(' : Result := ')';
    '[' : Result := ']';
    '{' : Result := '}';
  else
    Result := #0;
  end; {of case }
  end;


{ Scannerfunktionen fr logische Term-Bestandteile }

function N_Zahl(var s: String; var res: Double): Boolean;
  var c : Char;
  begin
  c := GetFirstChar(s);
  If c in ['0'..'9'] then begin  { Es wird eine Ziffer erwartet....    }
    Result := True;
    res := 0;
    Repeat                       { Solange Ziffern da sind, werden sie }
      Delete(s, 1, 1);
      res := 10 * res + (Ord(c) - Ord('0'));  { ... ausgewertet.       }
      c := GetFirstChar(s);                   { Nchstes Zeichen holen }
    until not (c in ['0'..'9']); { bis keine Ziffer mehr gefunden wird }
    end
  else begin                     { Falls keine Ziffer da ist: Fehler!  }
    Result := False;
    LastErrorMsg := 'Ziffer erwartet!';       { Fehlermeldung absetzen }
    end;
  end;


function Zahl(var s: String; var res: Double): Boolean;
  var IsMinus,
      IsMinusExp : Boolean;
      c          : Char;
      d, e       : Double;
  begin
  IsMinus := False;              { Normalerweise erwarten wir was Positives! }
  c := GetFirstChar(s);
  If c in ['+', '-'] then begin        { Optionales Vorzeichen testen }
    Delete(s, 1, 1);
    IsMinus := c = '-';                { Gegebenenfalls auf Minus umschalten }
    end;
  If N_Zahl(s, res) then begin         { Vorkomma-Anteil holen }
    Result := True;                    { => Erfolg ! }
    c := GetFirstChar(s);
    If c in ['.', ','] then begin      { Nachkommastellen vorhanden ? }
      Delete(s, 1, 1);                 { Dezimaltrenner lschen }
      c := GetFirstChar(s);
      If c in ['0'..'9'] then begin    { Ziffern einlesen }
        d := 0.1;                      { Stellenwert initialisieren }
        Repeat
          Delete(s, 1, 1);
          res := res + d * (Ord(c) - Ord('0'));  { Wert berechnen }
          d   := d / 10;               { Stellenwert fr nchste Stelle }
          c   := GetFirstChar(s);      { Nchste Ziffer holen ... }
        until not (c in ['0'..'9']);   { ...bis keine mehr da sind }
        end
      else begin
        Result := False;
        LastErrorMsg := 'Nachkommastellen erwartet!';
        end;
      end;
    If c in ['e', 'E'] then begin      { Exponent vorhanden ? }
      Delete(s, 1, 1);                 { 'e' lschen }
      IsMinusExp := False;
      c := GetFirstChar(s);
      If c in ['+', '-'] then begin    { optionales Vorzeichen testen }
        Delete(s, 1, 1);
        IsMinusExp := c = '-';         { gegebenenfalls auf Minus umschalten }
        end;
      If N_Zahl(s, e) then begin       { Exponent holen }
        If IsMinusExp then e := -e;
        res := res * power(10, e);
        end
      else begin                       { Kein Exponent vorhanden ? => Fehler! }
        Result := False;
        LastErrorMsg := 'Exponent erwartet!';
        end;
      end;
    If IsMinus then res := -res;       { Erst hier das VZ einrechnen !!!      }
    end
  else
    Result := False;            { LastErrorMsg wird schon von N_Zahl gesetzt. }
  end;


function Summe(var s: String; var res: Double): Boolean; forward;


function Klammerterm(var s: String; var res: Double): Boolean;
  var c1, c2 : Char;
  begin
  c1 := GetFirstChar(s);
  If c1 in ['(', '[', '{'] then begin  { Klammerterm }
    Delete(s, 1, 1);
    If Summe(s, res) then begin        { Klammerinhalt berechnen }
      c2 := GetFirstChar(s);
      If c2 = ClosingBracket(c1) then begin
        Delete(s, 1, 1);
        Result := True;
        end
      else begin
        Result := False;
        LastErrorMsg := '"' + ClosingBracket(c1)+ '" erwartet!';
        end;
      end
    else
      Result := False;                 { Fehlermeldung bernehmen }
    end
  else
    Result := False;
  end;


function VarKonst(var s: String; var res: Double): Boolean;
  var c1, c2 : Char;
  begin
  Result := True;     { Optimistische Annahme, die Quelltext sparen hilft }
  c1 := Upcase(GetFirstChar(s));
  Case c1 of
    'X' : begin
          Delete(s, 1, 1);
          res := arg_x;
          end;
    'E' : begin
          Delete(s, 1, 1);
          res := const_e;
          end;
    'P' : begin
          Delete(s, 1, 1);
          c2 := Upcase(GetFirstChar(s));
          If c2 = 'I' then begin
            Delete(s, 1, 1);
            res := Pi;
            end
          else
            Result := False;
          end;
  else
    Result := False;
  end; { of case }
  end;


function VFaktor(var s: String; var res: Double; var HasMinus: Boolean): Boolean;
  var c : Char;
  begin
  HasMinus := False;              { Normalerweise erwarten wir was Positives! }
  c := GetFirstChar(s);
  If c in ['+', '-'] then begin      { Optionales Vorzeichen testen         }
    Delete(s, 1, 1);
    HasMinus := c = '-';             { Gegebenenfalls auf Minus umschalten  }
    end;
  Result := Klammerterm(s, res) or   { Jede dieser Funktionen wird nur dann }
            Zahl(s, res) or          { wirklich aufgerufen, wenn alle zuvor }
            VarKonst(s, res);        { aufgerufenen FALSE geliefert haben.  }
  If HasMinus then                    { Vorzeichen erst hier einrechnen !    }
    res := -res;
  If Result then
    LastErrorMsg := '';              { Eventuelle Fehlermeldung lschen     }
  end;


function Potenz(var s: String; var res: Double): Boolean;
  var BaseHasMinus,
      ExpHasMinus : Boolean;
      r1 : Double;
      c  : Char;
  begin
  If VFaktor(s, res, BaseHasMinus) then begin     { Basis holen }
    Result := True;
    c := GetFirstChar(s);
    While c = '^' do begin       { Falls Exponent vorhanden ist: }
      Delete(s, 1, 1);             { Hochzeichen lschen und...  }
      If VFaktor(s, r1, ExpHasMinus) then begin { Exponent holen }
        If BaseHasMinus then
          res := - power(-res, r1)  { Damit -3^2 < 0 wird ! }
        else
          res := power(res, r1);
        c := GetFirstChar(s);
        end
      else begin
        Result := False;          { Fehlermeldung bernehmen    }
        c := #0;
        end;
      end;
    end                         { Kein Exponent => nix zu tun ! }
  else
    Result := False;            { Fehlermeldung bernehmen      }
  end;


function Produkt(var s: String; var res: Double): Boolean;
  var r1 : Double;
      c  : Char;
  begin
  If Potenz(s, res) then begin     { Ersten Faktor holen                     }
    Result := True;
    c := GetFirstChar(s);
    While c in ['*', '/'] do begin { Solange eine Punktrechnung folgt :      }
      Delete(s, 1, 1);                 { Rechenzeichen lschen               }
      If Potenz(s, r1) then begin      { Nchsten Faktor holen...            }
        If c = '*' then
          res := res * r1              { und zum Produkt multiplizieren .... }
        else
          res := res / r1;             { ...bzw. dividieren                  }
        c := GetFirstChar(s);
        end
      else begin                       { Kein Faktor vorhanden ? => Fehler!  }
        Result := False;               {  LastErrorMsg wird in Zahl gesetzt  }
        c      := #0;                  { While-Schleife verlassen }
        end;
      end;
    end
  else                             { Kein Faktor da ? => Fehler!             }
    Result := False;               {     LastErrorMsg wird in Zahl gesetzt   }
  end;


function Summe {(var s: String; var res: Double): Boolean};
  var r1 : Double;
      c  : Char;
  begin
  If Produkt(s, res) then begin    { Ersten Summanden holen                 }
    Result := True;
    c := GetFirstChar(s);
    While c in ['+', '-'] do begin { Solange eine Strichrechnung folgt :    }
      Delete(s, 1, 1);             { Rechenzeichen lschen                  }
      If Produkt(s, r1) then begin { Nchsten Summanden holen...            }
        If c = '+' then
          res := res + r1          { und zur aktuellen Summe addieren...    }
        else
          res := res - r1;         { ...bzw. von ihr subtrahieren           }
        c := GetFirstChar(s);
        end
      else begin                   { Kein Summand vorhanden ? => Fehler!    }
        Result := False;           {  LastErrorMsg wird in Produkt gesetzt  }
        c      := #0;              { While-Schleife verlassen }
        end;
      end;
    end
  else                             { Kein Summand da ? => Fehler!           }
    Result := False;               {  LastErrorMsg wird in Produkt gesetzt  }
  end;


{ Interface-Funktionen (ffentlich) }

function TermWert(s: String; x: Double): Double;
  { Diese Funktion dient hauptschlich der Fehlerverwaltung. Sie sorgt dafr,
    dass bei einem Fehler whrend des Interpretationsvorgangs die Variablen
    LastErrorMsg und LastErrorPos korrekt gesetzt werden. Auerdem verwaltet
    sie die Variable Okay, in der ber den Erfolg des Interpretationsversuchs
    Buch gefhrt wird.
    Im Gegensatz zu den internen Interpreter-Funktionen  }
  var r   : Double;
      ils : Integer;
  begin
  Okay  := True;
  arg_x := x;
  ils   := Length(s);
  LastErrorMsg := '';
  LastErrorPos := ils;
  IF Summe(s, r) then begin  { Hier wird gerechnet: jeder Term ist eine Summe. }
    Result := r;
    Okay   := True;
    end
  else begin                    { Buchhaltung im Fehlerfall: }
    Result := 0;
    Okay := False;
    LastErrorPos := ils - Length(s);
    LastErrorMsg := LastErrorMsg + '  {Rest : ' + s + ' }';
    end;
  end;

function IsOkay: Boolean;
  begin
  Result := Okay;
  end;

function GetLastErrorMsg: String;
  begin
  Result := LastErrorMsg;
  end;

function GetLastErrorPos: Integer;
  begin
  Result := LastErrorPos;
  end;

initialization
  LastErrorMsg := '';
  Okay     := True;
  const_e  := exp(1);
end.
